var _a;
import { Account, CoValueBase, Ref, SchemaInit, inspect, makeRefs, subscriptionsScopes, ItemsSym, isRefEncoded, loadCoValue, subscribeToCoValue, ensureCoValueLoaded, subscribeToExistingCoValue, } from "../internal.js";
/**
 * CoMaps are collaborative versions of plain objects, mapping string-like keys to values.
 *
 * @categoryDescription Declaration
 * Declare your own CoMap schemas by subclassing `CoMap` and assigning field schemas with `co`.
 *
 * Optional `co.ref(...)` fields must be marked with `{ optional: true }`.
 *
 * ```ts
 * import { co, CoMap } from "jazz-tools";
 *
 * class Person extends CoMap {
 *   name = co.string;
 *   age = co.number;
 *   pet = co.ref(Animal);
 *   car = co.ref(Car, { optional: true });
 * }
 * ```
 *
 * @categoryDescription Content
 * You can access properties you declare on a `CoMap` (using `co`) as if they were normal properties on a plain object, using dot notation, `Object.keys()`, etc.
 *
 * ```ts
 * person.name;
 * person["age"];
 * person.age = 42;
 * person.pet?.name;
 * Object.keys(person);
 * // => ["name", "age", "pet"]
 * ```
 *
 * @category CoValues
 *  */
export class CoMap extends CoValueBase {
    /** @internal */
    get _schema() {
        return this.constructor._schema;
    }
    /**
     * If property `prop` is a `co.ref(...)`, you can use `coMaps._refs.prop` to access
     * the `Ref` instead of the potentially loaded/null value.
     *
     * This allows you to always get the ID or load the value manually.
     *
     * @example
     * ```ts
     * person._refs.pet.id; // => ID<Animal>
     * person._refs.pet.value;
     * // => Animal | null
     * const pet = await person._refs.pet.load();
     * ```
     *
     * @category Content
     **/
    get _refs() {
        return makeRefs((key) => this._raw.get(key), () => {
            const keys = this._raw.keys().filter((key) => {
                const schema = this._schema[key] ||
                    this._schema[ItemsSym];
                return schema && schema !== "json" && isRefEncoded(schema);
            });
            return keys;
        }, this._loadedAs, (key) => (this._schema[key] ||
            this._schema[ItemsSym]));
    }
    /** @category Collaboration */
    get _edits() {
        return new Proxy(this, {
            get(target, key) {
                const rawEdit = target._raw.lastEditAt(key);
                if (!rawEdit)
                    return undefined;
                const descriptor = target._schema[key];
                return {
                    value: descriptor === "json"
                        ? rawEdit.value
                        : "encoded" in descriptor
                            ? descriptor.encoded.encode(rawEdit.value)
                            : new Ref(rawEdit.value, target._loadedAs, descriptor).accessFrom(target, "_edits." + key.toString() + ".value"),
                    ref: descriptor !== "json" && isRefEncoded(descriptor)
                        ? new Ref(rawEdit.value, target._loadedAs, descriptor)
                        : undefined,
                    by: rawEdit.by &&
                        new Ref(rawEdit.by, target._loadedAs, {
                            ref: Account,
                            optional: false,
                        }).accessFrom(target, "_edits." + key.toString() + ".by"),
                    madeAt: rawEdit.at,
                };
            },
        });
    }
    /** @internal */
    get _loadedAs() {
        return Account.fromNode(this._raw.core.node);
    }
    /** @internal */
    constructor(
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    options) {
        super();
        if (options) {
            if ("fromRaw" in options) {
                Object.defineProperties(this, {
                    id: {
                        value: options.fromRaw.id,
                        enumerable: false,
                    },
                    _raw: { value: options.fromRaw, enumerable: false },
                });
            }
            else {
                throw new Error("Invalid CoMap constructor arguments");
            }
        }
        return new Proxy(this, CoMapProxyHandler);
    }
    /**
     * Create a new CoMap with the given initial values and owner.
     *
     * The owner (a Group or Account) determines access rights to the CoMap.
     *
     * The CoMap will immediately be persisted and synced to connected peers.
     *
     * @example
     * ```ts
     * const person = Person.create({
     *   name: "Alice",
     *   age: 42,
     *   pet: cat,
     * }, { owner: friendGroup });
     * ```
     *
     * @category Creation
     **/
    static create(init, options) {
        const instance = new this();
        const raw = instance.rawFromInit(init, options.owner);
        Object.defineProperties(instance, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
        return instance;
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    toJSON(_key, seenAbove) {
        const jsonedFields = this._raw.keys().map((key) => {
            const tKey = key;
            const descriptor = (this._schema[tKey] ||
                this._schema[ItemsSym]);
            if (descriptor == "json" || "encode" in descriptor) {
                return [key, this._raw.get(key)];
            }
            else if (isRefEncoded(descriptor)) {
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                if (seenAbove?.includes(this[tKey]?.id)) {
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    return [key, { _circular: this[tKey]?.id }];
                }
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                const jsonedRef = this[tKey]?.toJSON(tKey, [
                    ...(seenAbove || []),
                    this.id,
                ]);
                return [key, jsonedRef];
            }
            else {
                return [key, undefined];
            }
        });
        return {
            id: this.id,
            _type: this._type,
            ...Object.fromEntries(jsonedFields),
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    /** @internal */
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    rawFromInit(init, owner) {
        const rawOwner = owner._raw;
        const rawInit = {};
        if (init)
            for (const key of Object.keys(init)) {
                const initValue = init[key];
                const descriptor = (this._schema[key] || this._schema[ItemsSym]);
                if (!descriptor) {
                    continue;
                }
                if (descriptor === "json") {
                    rawInit[key] = initValue;
                }
                else if (isRefEncoded(descriptor)) {
                    if (initValue) {
                        rawInit[key] = initValue.id;
                    }
                }
                else if ("encoded" in descriptor) {
                    rawInit[key] = descriptor.encoded.encode(
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    initValue);
                }
            }
        return rawOwner.createMap(rawInit);
    }
    /**
     * Declare a Record-like CoMap schema, by extending `CoMap.Record(...)` and passing the value schema using `co`. Keys are always `string`.
     *
     * @example
     * ```ts
     * import { co, CoMap } from "jazz-tools";
     *
     * class ColorToFruitMap extends CoMap.Record(
     *  co.ref(Fruit)
     * ) {}
     *
     * // assume we have map: ColorToFruitMap
     * // and strawberry: Fruit
     * map["red"] = strawberry;
     * ```
     *
     * @category Declaration
     */
    static Record(value) {
        var _b;
        // eslint-disable-next-line @typescript-eslint/no-unsafe-declaration-merging
        class RecordLikeCoMap extends _a {
            constructor() {
                super(...arguments);
                this[_b] = value;
            }
        }
        _b = ItemsSym;
        return RecordLikeCoMap;
    }
    /**
     * Load a `CoMap` with a given ID, as a given account.
     *
     * `depth` specifies which (if any) fields that reference other CoValues to load as well before resolving.
     * The `DeeplyLoaded` return type guarantees that corresponding referenced CoValues are loaded to the specified depth.
     *
     * You can pass `[]` or `{}` for shallowly loading only this CoMap, or `{ fieldA: depthA, fieldB: depthB }` for recursively loading referenced CoValues.
     *
     * Check out the `load` methods on `CoMap`/`CoList`/`CoStream`/`Group`/`Account` to see which depth structures are valid to nest.
     *
     * @example
     * ```ts
     * const person = await Person.load(
     *   "co_zdsMhHtfG6VNKt7RqPUPvUtN2Ax",
     *   me,
     *   { pet: {} }
     * );
     * ```
     *
     * @category Subscription & Loading
     */
    static load(id, as, depth) {
        return loadCoValue(this, id, as, depth);
    }
    /**
     * Load and subscribe to a `CoMap` with a given ID, as a given account.
     *
     * Automatically also subscribes to updates to all referenced/nested CoValues as soon as they are accessed in the listener.
     *
     * `depth` specifies which (if any) fields that reference other CoValues to load as well before calling `listener` for the first time.
     * The `DeeplyLoaded` return type guarantees that corresponding referenced CoValues are loaded to the specified depth.
     *
     * You can pass `[]` or `{}` for shallowly loading only this CoMap, or `{ fieldA: depthA, fieldB: depthB }` for recursively loading referenced CoValues.
     *
     * Check out the `load` methods on `CoMap`/`CoList`/`CoStream`/`Group`/`Account` to see which depth structures are valid to nest.
     *
     * Returns an unsubscribe function that you should call when you no longer need updates.
     *
     * Also see the `useCoState` hook to reactively subscribe to a CoValue in a React component.
     *
     * @example
     * ```ts
     * const unsub = Person.subscribe(
     *   "co_zdsMhHtfG6VNKt7RqPUPvUtN2Ax",
     *   me,
     *   { pet: {} },
     *   (person) => console.log(person)
     * );
     * ```
     *
     * @category Subscription & Loading
     */
    static subscribe(id, as, depth, listener) {
        return subscribeToCoValue(this, id, as, depth, listener);
    }
    /**
     * Given an already loaded `CoMap`, ensure that the specified fields are loaded to the specified depth.
     *
     * Works like `CoMap.load()`, but you don't need to pass the ID or the account to load as again.
     *
     * @category Subscription & Loading
     */
    ensureLoaded(depth) {
        return ensureCoValueLoaded(this, depth);
    }
    /**
     * Given an already loaded `CoMap`, subscribe to updates to the `CoMap` and ensure that the specified fields are loaded to the specified depth.
     *
     * Works like `CoMap.subscribe()`, but you don't need to pass the ID or the account to load as again.
     *
     * Returns an unsubscribe function that you should call when you no longer need updates.
     *
     * @category Subscription & Loading
     **/
    subscribe(depth, listener) {
        return subscribeToExistingCoValue(this, depth, listener);
    }
    applyDiff(newValues) {
        for (const key in newValues) {
            if (Object.prototype.hasOwnProperty.call(newValues, key)) {
                const tKey = key;
                const descriptor = (this._schema[tKey] ||
                    this._schema[ItemsSym]);
                if (tKey in this._schema) {
                    const newValue = newValues[tKey];
                    const currentValue = this[tKey];
                    if (descriptor === "json" || "encoded" in descriptor) {
                        if (currentValue !== newValue) {
                            // eslint-disable-next-line @typescript-eslint/no-explicit-any
                            this[tKey] = newValue;
                        }
                    }
                    else if (isRefEncoded(descriptor)) {
                        const currentId = currentValue
                            ?.id;
                        const newId = newValue?.id;
                        if (currentId !== newId) {
                            // eslint-disable-next-line @typescript-eslint/no-explicit-any
                            this[tKey] = newValue;
                        }
                    }
                }
            }
        }
        return this;
    }
}
_a = CoMap;
(() => {
    _a.prototype._type = "CoMap";
})();
// TODO: cache handlers per descriptor for performance?
const CoMapProxyHandler = {
    get(target, key, receiver) {
        if (key === "_schema") {
            return Reflect.get(target, key);
        }
        else if (key in target) {
            return Reflect.get(target, key, receiver);
        }
        else {
            const descriptor = (target._schema[key] ||
                target._schema[ItemsSym]);
            if (descriptor && typeof key === "string") {
                const raw = target._raw.get(key);
                if (descriptor === "json") {
                    return raw;
                }
                else if ("encoded" in descriptor) {
                    return raw === undefined
                        ? undefined
                        : descriptor.encoded.decode(raw);
                }
                else if (isRefEncoded(descriptor)) {
                    return raw === undefined
                        ? undefined
                        : new Ref(raw, target._loadedAs, descriptor).accessFrom(receiver, key);
                }
            }
            else {
                return undefined;
            }
        }
    },
    set(target, key, value, receiver) {
        var _b;
        if ((typeof key === "string" || ItemsSym) &&
            typeof value === "object" &&
            value !== null &&
            SchemaInit in value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[key] =
                value[SchemaInit];
            return true;
        }
        const descriptor = (target._schema[key] ||
            target._schema[ItemsSym]);
        if (descriptor && typeof key === "string") {
            if (descriptor === "json") {
                target._raw.set(key, value);
            }
            else if ("encoded" in descriptor) {
                target._raw.set(key, descriptor.encoded.encode(value));
            }
            else if (isRefEncoded(descriptor)) {
                target._raw.set(key, value.id);
                subscriptionsScopes
                    .get(target)
                    ?.onRefAccessedOrSet(target.id, value.id);
            }
            return true;
        }
        else {
            return Reflect.set(target, key, value, receiver);
        }
    },
    defineProperty(target, key, attributes) {
        var _b;
        if ("value" in attributes &&
            typeof attributes.value === "object" &&
            SchemaInit in attributes.value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[key] =
                attributes.value[SchemaInit];
            return true;
        }
        else {
            return Reflect.defineProperty(target, key, attributes);
        }
    },
    ownKeys(target) {
        const keys = Reflect.ownKeys(target).filter((k) => k !== ItemsSym);
        // for (const key of Reflect.ownKeys(target._schema)) {
        //     if (key !== ItemsSym && !keys.includes(key)) {
        //         keys.push(key);
        //     }
        // }
        for (const key of target._raw.keys()) {
            if (!keys.includes(key)) {
                keys.push(key);
            }
        }
        return keys;
    },
    getOwnPropertyDescriptor(target, key) {
        if (key in target) {
            return Reflect.getOwnPropertyDescriptor(target, key);
        }
        else {
            const descriptor = (target._schema[key] ||
                target._schema[ItemsSym]);
            if (descriptor || key in target._raw.ops) {
                return {
                    enumerable: true,
                    configurable: true,
                    writable: true,
                };
            }
        }
    },
    has(target, key) {
        const descriptor = (target._schema?.[key] ||
            target._schema?.[ItemsSym]);
        if (target._raw && typeof key === "string" && descriptor) {
            return target._raw.get(key) !== undefined;
        }
        else {
            return Reflect.has(target, key);
        }
    },
    deleteProperty(target, key) {
        const descriptor = (target._schema[key] ||
            target._schema[ItemsSym]);
        if (typeof key === "string" && descriptor) {
            target._raw.delete(key);
            return true;
        }
        else {
            return Reflect.deleteProperty(target, key);
        }
    },
};
//# sourceMappingURL=coMap.js.map